import requests
import json
from ...util import log

class HPLoadBalancer():
    """
    Convenience functions to manage HP cloud LBaaS instances. 
    LBaaS uses its own AZ-independent public URL for management,
    but accepts the common auth token. Rather than integrating with
    openstack, this is HP-cloud specific for now, since there is still
    disagreement about the direction openstack's load balancing effort
    will take.

    TODO: The beta api doesn't seem to support tags, which we use elsewhere
    for filtering a stack. Instead this'll use 'name'
    """
    def __init__(self, hpcloud):
        """
        Provide a management URL (from the openstack service catalog)
        and auth token (which can be pinched from novaclient)
        """

        lb_catalog = filter(lambda c: c['name'] == 'Load Balancer', 
                                   hpcloud.os_catalog['access']['serviceCatalog'])
        lb_management_url = lb_catalog[0]['endpoints'][0]['publicURL']

        self.management_url = lb_management_url
        self.auth_token = hpcloud.os_auth_token

    def list_lbs(self):
        """
        Lists all LBaaS instances. To get details (nodes, external IPs)
        use the ``lb_details`` call.

        :rtype :class:`list` of :class:`dict`
        """
        resp, body = self._request('get', '/loadbalancers')
        return body['loadBalancers']

    def lb_details(self, lb_id):
        """
        Get details (including all nodes and external IPs)

        :param string lb_id:  Load balancer id

        :rtype :class:`dict`
        """
        resp, body = self._request('get', '/loadbalancers/%s' % lb_id)
        return body

    def find_lb_by_name(self, name):
        """
        Look up a LBaaS instance by name (rather than id)

        :attr string name:  The LBaaS name assigned at creation time

        :rtype :class:`dict`
        """
        log.debug("Finding load balancers matching name '%s'" % name)
        matching = filter(lambda l: l['name'] == name, self.list_lbs())
        if len(matching) > 1:
            raise ValueError("Ambiguous; more than one load balancer matched '%s'" % name)
        if matching:
            log.info("Found existing load balancer, %s" % matching[0]['id'])
            return matching[0]
        return None

    def create_lb(self, name, protocol='HTTP', port=80, algorithm=None,
            virtual_ips=[], nodes=[], node_port=None):
        """
        Create a new LBaaS instance. A name is required. If no nodes
        are required, the instance ``status`` (after it's built) will be
        ``ERROR`` until nodes are added. The return dict contains the id
        and ``virtualIps`` (the external IP).

        :param string name:  The LBaaS name - should be unique

        :param string protocol:  Supported: HTTP, TCP

        :param int port:  The external port. Supported: 80, 443, seemingly

        :param list virtual_ips:  Use an existing IP (allows multiple
                                  ports per IP)

        :param list nodes:  Nodes addresses to add
        """
        log.info("Creating load balancer '%s'" % name)
        protocol = protocol.upper()
        nodes_to_add = []
        if nodes:
            nodes_to_add = map(
                    lambda n: {'address': n, 'port': str(node_port)}, 
                    nodes
                    )
        data = {
            'name': name,
            'protocol': protocol,
            'port': str(port),
            'nodes': nodes_to_add,
        }
        if virtual_ips:
            data['virtualIps'] = map(lambda i: {'id': i}, virtual_ips)
        if algorithm:
            data['algorithm'] = algorithm
        resp, body = self._request('post', '/loadbalancers', data=data)
        return body

    def delete_lb(self, lb_id):
        """
        Delete an instance

        :param string lb_id:  Delete this LBaaS id
        """
        log.info("Deleting load balancer %s" % lb_id)
        self._request('delete', '/loadbalancers/%s' % lb_id)

    def add_lb_nodes(self, lb_id, nodes):
        """
        Adds nodes to an existing LBaaS instance

        :param string lb_id:  Balancer id

        :param list nodes:  Nodes to add. {address, port, [condition]}

        :rtype :class:`list`
        """
        log.info("Adding load balancer nodes %s" % nodes)
        resp, body = self._request(
                'post', 
                '/loadbalancers/%s/nodes' % lb_id,
                data={'nodes': nodes})
        return body

    def match_lb_nodes(self, lb_id, existing_nodes, host_addresses, host_port):
        """
        Add and remove nodes to match the host addresses
        and port given, based on existing_nodes

        :param string lb_id: Load balancer id

        :param :class:`list` of :class:`dict` existing_nodes: Existing nodes

        :param :class:`list` host_addresses: Node host addresses

        :param string port: Node port
        """
        delete_filter = lambda n: \
                n['address'] not in host_addresses or \
                str(n['port']) != str(host_port)

        delete_nodes = filter(delete_filter, existing_nodes)
        delete_node_ids = [n['id'] for n in delete_nodes]
        delete_node_hosts = [n['address'] for n in delete_nodes]
        
        if delete_node_ids:
            args = (lb_id, delete_node_ids)
            self.remove_lb_nodes(*args)

        current_nodes = set([n['address'] for n in existing_nodes])
        current_nodes -= set(delete_node_hosts)
        add_nodes = host_addresses - current_nodes
        
        if add_nodes:
            nodes_to_add = [
                    {'address': n, 'port': str(host_port)}
                    for n in add_nodes
            ]
            args = (lb_id, nodes_to_add)
            self.add_lb_nodes(*args)
        
        log.info("Were %d nodes. Added %d nodes; deleted %d nodes" % 
                (len(existing_nodes), len(add_nodes), len(delete_nodes)))

    def remove_lb_nodes(self, lb_id, node_ids):
        """
        Remove one or more nodes

        :param string lb_id:  Balancer id

        :param list node_ids:  List of node ids
        """
        log.info("Removing load balancer nodes %s" % node_ids)
        for node_id in node_ids:
            self._request('delete', '/loadbalancers/%s/nodes/%s' % (lb_id, node_id))

    def update_lb_node_condition(self, lb_id, node_id, condition):
        """
        Update node condition - specifically to disable/enable

        :param string lb_id:  Balancer id

        :param string node_id:  Node id

        :param string condition:  ENABLED/DISABLED
        """
        self._request(
                'put', 
                '/loadbalancers/%s/nodes/%s' % (lb_id, node_id),
                data={'condition': condition})
        
    def _request(self, method, url, data=None, **kwargs):
        kwargs.setdefault('headers', {})['X-Auth-Token'] = self.auth_token
        kwargs.setdefault('headers', {})['Content-Type'] = 'application/json'
        if data:
            if isinstance(data, dict):
                kwargs['data'] = json.dumps(data)
            else:
                kwargs['data'] = data

        url = '%s%s' % (self.management_url, url)
        resp = requests.request(method, url, **kwargs)
        if resp.text:
            try:
                body = json.loads(resp.text)
            except ValueError:
                body = None
        else:
            body = None
        resp.raise_for_status()
        return resp, body




# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
import boto
import boto.ec2
from boto.exception import EC2ResponseError
import time

from .. import BangError, TimeoutError, resources as R, attributes as A
from ..util import log, poll_with_timeout
from .bases import Provider, Consul


DEFAULT_TIMEOUT_S = 120


def server_to_dict(server):
    """
    Returns the :class:`dict` representation of a server object.

    The returned :class:`dict` is meant to be consumed by
    :class:`~bang.deployers.cloud.ServerDeployer` objects.

    """
    return {
            A.server.ID: server.id,
            A.server.PUBLIC_IPS: [server.public_dns_name],
            A.server.PRIVATE_IPS: [server.private_dns_name],
            }


class EC2SecGroup(object):
    """
    Represents an EC2 security group.

    The :attr:`rules` attribute is a specialized dict whose keys are the
    *normalized* rule definitions, and whose values are EC2 grants which can be
    kwargs-expanded when passing
    :meth:`boto.ec2.securitygroup.SecurityGroup.revoke`.  E.g.::

        {
            ('tcp', 1, 65535, 'group-foo'): {
                'ip_protocol': 'tcp',
                'from_port': '1',
                'to_port': '65535',
                'src_group': 'group-foo',
                'target': SecurityGroup:group-bar,
                },
            ('tcp', 8080, 8080, '15.183.202.114/32'):  {
                'ip_protocol': 'tcp',
                'from_port': '8080',
                'to_port': '8080',
                'cidr_ip': '15.183.202.114/32',
                'target': SecurityGroup:group-bar,
                },
        }

    This also maintains a reference to the original
    :class:`boto.ec2.securitygroup.SecurityGroup` instance.

    Suitable for returning from :meth:`EC2.find_secgroup`.

    """
    def __init__(self, ec2sg):
        owner_id = ec2sg.owner_id
        rules = {}
        for rule in ec2sg.rules:
            p = rule.ip_protocol
            f = int(rule.from_port)
            t = int(rule.to_port)
            core = {
                    'ip_protocol': p,
                    'from_port': f,
                    'to_port': t,
                    'target': ec2sg,
                    }
            for g in rule.grants:
                parsed = {}
                if g.cidr_ip:
                    s = parsed['source'] = str(g.cidr_ip)
                elif g.owner_id == owner_id and g.name == ec2sg.name:
                    parsed['source_self'] = True
                    s = ec2sg.name
                else:
                    parsed['source_group'] = '%s/%s' % (g.owner_id, g.name)
                    s = g.name
                parsed.update(core)
                rules[(p, f, t, s)] = parsed
        self.rules = rules
        self.ec2sg = ec2sg


class EC2(Consul):
    """The consul for the compute service in AWS (EC2)."""
    def __init__(self, *args, **kwargs):
        super(EC2, self).__init__(*args, **kwargs)
        creds = self.provider.creds
        self.access_key_id = creds[A.creds.ACCESS_KEY_ID]
        self.secret_key = creds[A.creds.SECRET_ACCESS_KEY]
        self._ec2 = None

    @property
    def ec2(self):
        if not self._ec2:
            # this connection lets boto pick the default region.  be sure to use
            # set_region() if you need a specific region.
            self._ec2 = boto.connect_ec2(self.access_key_id, self.secret_key)
        return self._ec2

    def set_region(self, region_name):
        log.debug("Setting region to %s" % region_name)
        self._ec2 = boto.ec2.connect_to_region(
                region_name,
                aws_access_key_id=self.access_key_id,
                aws_secret_access_key=self.secret_key,
                )

    def find_servers(self, tags, running=True):
        """
        Returns any servers in the region that have tags that match the
        key-value pairs in :attr:`tags`.

        :param Mapping tags:  A mapping object in which the keys are the tag
            names and the values are the tag values.

        :param bool running:  A flag to limit server list to instances that are
            actually *running*.

        :rtype:  :class:`list` of :class:`dict` objects.  Each :class:`dict`
            describes a single server instance.

        """
        filters = dict([('tag:%s' % key, val) for key, val in tags.items()])
        if running:
            filters['instance-state-name'] = 'running'

        res = self.ec2.get_all_instances(filters=filters)
        instances = [server_to_dict(i) for r in res for i in r.instances]
        log.debug('instances: %s' % instances)
        return instances

    def create_server(self, basename, disk_image_id, instance_type,
            ssh_key_name, tags=None, availability_zone=None,
            timeout_s=DEFAULT_TIMEOUT_S, **kwargs):
        """
        Creates a new server instance.  This call blocks until the server is
        created and available for normal use, or :attr:`timeout_s` has elapsed.

        :param str basename:  An identifier for the server.  A random postfix
            will be appended to this basename to work around OpenStack Nova
            REST API limitations.

        :param str disk_image_id:  The identifier of the base disk image to use
            as the rootfs.

        :param str instance_type:  The name of an OpenStack instance type, or
            *flavor*.  This is specific to the OpenStack provider installation.

        :param str ssh_key_name:  The name of the ssh key to inject into the
            target server's ``authorized_keys`` file.  The key must already
            have been registered with the OpenStack Nova provider.

        :param tags:  Up to 5 key-value pairs of arbitrary strings to use as
            *tags* for the server instance.
        :type tags:  :class:`Mapping`

        :param str availability_zone:  The name of the availability zone in
            which to place the server.

        :param float timeout_s:  The number of seconds to poll for an active
            server before failing.  Defaults to ``0`` (i.e. Expect server to be
            active immediately).

        :rtype:  :class:`dict`

        """
        log.info(
                'Launching server %s... this could take a while...'
                % basename
                )
        res = self.ec2.run_instances(
                disk_image_id,
                instance_type=instance_type,
                key_name=ssh_key_name,
                placement=availability_zone,
                disable_api_termination=True,
                **kwargs
                )
        instance = res.instances[0]

        # we're too fast for EC2... slow down a little bit, twice
        time.sleep(2)

        def apply_tags():
            try:
                for key, val in tags.items():
                    instance.add_tag(key, val)
                return True
            except EC2ResponseError:
                pass
        if not poll_with_timeout(timeout_s, apply_tags, 5):
            raise TimeoutError('Could not tag server %s' % instance.id)

        def find_running_instance():
            if instance.update() == 'running':
                return instance
        running = poll_with_timeout(timeout_s, find_running_instance, 5)
        if not running:
            raise TimeoutError('Could not launch server within allotted time.')
        return server_to_dict(running)

    def find_secgroup(self, name):
        """
        Find a security group by name.

        Returns a :class:`EC2SecGroup` instance if found, otherwise returns
        None.

        """
        res = self.ec2.get_all_security_groups(filters={'group-name': name})
        if res:
            return EC2SecGroup(res[0])

    def create_secgroup(self, name, description):
        """
        Creates a new server security group.

        :param str name:  The name of the security group to create.
        :param str description:  A short description of the group.

        """
        return self.ec2.create_security_group(name, description)
        log.debug("... created group %s" % name)

    def create_secgroup_rule(self, protocol, from_port, to_port,
            source, target):
        """
        Creates a new server security group rule.

        :param str protocol:  E.g. ``tcp``, ``icmp``, etc...
        :param int from_port:  E.g. ``1``
        :param int to_port:  E.g. ``65535``
        :param str source:
        :param str target:  The target security group.  I.e. the group in which
            this rule should be created.

        """
        kwargs = {
            'ip_protocol': protocol,
            'from_port': from_port,
            'to_port': to_port
        }
        sg = self.find_secgroup(target).ec2sg
        if not sg:
            raise BangError("Security group not found, %s" % target)
        if '/' in source:
            # Treat as cidr (/ is mask)
            kwargs['cidr_ip'] = source
        else:
            kwargs['src_group'] = self.find_secgroup(source).ec2sg
        sg.authorize(**kwargs)

    def delete_secgroup_rule(self, rule_def):
        """Deletes the security group rule identified by :attr:`rule_def`"""
        sg = rule_def.pop('target')
        sg.revoke(**rule_def)


class S3(Consul):
    pass


class RDS(Consul):
    pass


class AWS(Provider):

    CONSUL_MAP = {
            R.SERVERS: EC2,
            R.SERVER_SECURITY_GROUPS: EC2,
            R.SERVER_SECURITY_GROUP_RULES: EC2,
            R.BUCKETS: S3,
            R.DATABASES: RDS,
            }



# Copyright 2012 - John Calixto
#
# This file is part of bang.
#
# bang is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# bang is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with bang.  If not, see <http://www.gnu.org/licenses/>.
"""
Constants for attribute names of the various resources.

Just a place to stash more magic strings.
"""
from . import creds, server, secgroup, tags, database, ssh_key, loadbalancer

# stack attributes
NAME = 'name'
VERSION = 'version'
LOGGING = 'logging'
PLAYBOOKS = 'playbooks'
STACK = 'stack'
DEPLOYER_CREDS = 'deployer_credentials'
KEY = 'key'
SERVER_CLASS = 'server_class'
PROVIDER = 'provider'

# Like chicken fried chicken... this is a way to configure the name of the tag
# in which the combined stack-role (a.k.a. *name*) will be stored.  By default,
# unless this is specified directly in ~/.bangrc, the *name* value will be
# assigned to a tag named "Name" (this is the default tag displayed in the AWS
# management console).  I.e. using Bang defaults, the server named "bar" in the
# stack named "foo" will have the following tags:
#
#     stack:  foo
#     role:   bar
#     Name:   foo-bar
#
# In some cases, admins may have other purposes for the "Name" tag.  If
# ~/.bangrc were to have ``name_tag_name`` set to ``descriptor``, then the
# server described above would have the following tags:
#
#     stack:       foo
#     role:        bar
#     descriptor:  foo-bar
#
# To prevent Bang from assigning the *name* value to a tag, assign an empty
# string to the ``name_tag_name`` attribute in ~/.bangrc.
NAME_TAG_NAME = 'name_tag_name'
